/*
 * Ambient - the ultimate desktop
 * ------------------------------
 * Copyright 2001-2005 by David Gerber <zapek@morphos.net>
 * Copyright 2005-2006 Ambient Open Source Team
 * All Rights Reserved
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 *
 * $Id: fastcopy.c,v 1.11 2012/09/03 08:33:25 geit Exp $
 */

#include "ambient.h"

/* public */
#include <dos/dos.h>
#include <libraries/asyncio.h>
#include <proto/asyncio.h>
#include <proto/exec.h>
#include <proto/dos.h>

/* private */
#include "fastcopy.h"
#include "fileaction.h"
#include "methodstack.h"
#include "classes.h"
#include "prefs.h"
#include "file_func.h"
#include "threads.h"
#include "time_func.h"
#include "smartreq.h"
#include "examine64.h"

/************************************************************************/

#define COPY_CANCEL  0
#define COPY_PROCEED 1
#define COPY_SKIP    2

/************************************************************************/

APTR fastcopy_allocbuf( CONST_STRPTR from, CONST_STRPTR to, ULONG *startsize )
{
	ULONG size;

	THREAD;

	size = max( max( get_blocksize( from ), get_blocksize( to ) ), 512 );

	if( startsize )
	{
		*startsize = size;
	}

	#if USE_COPYIOPREFS
	switch( _conf( copy_mode ) )
	{
		case CM_FIXED:
			size = max( size, _conf( copy_buffersize ) );
			break;

		case CM_BLOCK:
			break;

		case CM_INCREMENTAL:
			size = max( size, _conf( copy_buffersize ) );
			break;
	}
	#else
	size = max( size, COPYIO_BUFFERSIZE );
	#endif
	D(COPY,bug("allocated buffer, size: %ld, startsize: %ld\n", size, startsize ? *startsize : 0));
	return( malloc ( size ) );
}

/************************************************************************/

void fastcopy_freebuf( APTR buf )
{
	ASSERT(buf);

	free( buf );
}

/************************************************************************/

static ULONG get_copybufsize( ULONG *ctx_bufsize )
{
	#if USE_COPYIOPREFS
	switch( _conf( copy_mode ) )
	{
		case CM_FIXED:
			*ctx_bufsize = _conf( copy_buffersize );
			return( *ctx_bufsize );

		case CM_BLOCK:
			return( *ctx_bufsize );

		case CM_INCREMENTAL:
			if( *ctx_bufsize < _conf( copy_buffersize ) )
			{
				*ctx_bufsize <<= 1;
				D(COPY,bug("buffer incremented to %ld\n", *ctx_bufsize));
			}
			return( *ctx_bufsize );
		
		#ifdef DEBUG
		default:
			PDB(("NYI!\n"));
			break;
		#endif
	}
	return( 0 );
	#else
	if( *ctx_bufsize < COPYIO_BUFFERSIZE )
	{
		*ctx_bufsize <<= 1;
		D(COPY,bug("buffer incremented to %ld\n", *ctx_bufsize));
	}
	return( *ctx_bufsize );
	#endif
}

/************************************************************************/

ULONG fastcopy( APTR obj, CONST_STRPTR from, CONST_STRPTR to, ULONG blocksize, ULONG clone, APTR buf, ULONG *time, ULONG cnt, QUAD * totaldone, ULONG *reqflags )
{
#if COPYIO_USE_ASYNCIO
	struct AsyncFile *af1, *af2;
#else
	BPTR f1, f2;
#endif
	LONG got, got2;
	ULONG own_alloc;
	ULONG error = FALSE;
	ULONG retval = FASTCOPY_FAILED;
	ULONG aborted = FALSE;
	ULONG fileincomplete = FALSE;

	THREAD;
	ASSERT(from);
	ASSERT(to);
	
	#ifdef DEBUG
	if( obj )
	{
		CHECKOBJECT( obj );
	}
	#endif

	D(COPY,bug("copying from <%s> to <%s>, blocksize: %ld, clone: %ld, buf: %p\n", from, to, blocksize, clone, buf));

	if( *reqflags & FILEACTION_REQUEST_CANCEL )
	{
		return( FASTCOPY_NOSOURCE );
	}

	if( isdir( to ) )
	{
		smartreq_request_sync( NULL, "Ambient  Copy", "*_Ok", MV_Notification_Warning, "Destination '%s' is a directory.", to );
		return( FASTCOPY_NOSOURCE );
	}

	if( buf )
	{
		own_alloc = FALSE;
	} else {
		buf = fastcopy_allocbuf( from, to, &blocksize );
		own_alloc = TRUE;
	}

	if( buf )
	{
		D(COPY,bug("buffer allocated\n"));
		#if COPYIO_USE_ASYNCIO
		#if USE_COPYIOPREFS
		if ( ( af1 = OpenAsync( (STRPTR) from, MODE_READ, _conf( copy_readahead_buffersize ) ) ) )
		#else
		if ( ( af1 = OpenAsync( (STRPTR) from, MODE_READ, COPYIO_READAHEAD_BUFFERSIZE ) ) )
		#endif
		#else
		if ( ( f1 = Open( from, MODE_OLDFILE ) ) )
		#endif
		{  
			#if !COPYIO_USE_ASYNCIO
			LONG accessMode = MODE_NEWFILE;
			#endif
			ULONG res = COPY_PROCEED;

			if( exists( to ) )
			{
				ULONG mask = 0;
				ULONG deleteprotected = 0;
				ULONG writeprotected  = 0;

				#if !COPYIO_USE_ASYNCIO
				accessMode = MODE_OLDFILE;
				#endif

				/* XXX : as in dopus, check if a both files have a version and compare them (date, md5sum ) */

				switch( fileaction_replace( from, to, FILEACTION_REPLACE_COPY, reqflags ) )
				{
					case FILEACTION_PROCEED:
						res = COPY_PROCEED;
						break;
					case FILEACTION_SKIP:
						res = COPY_SKIP;
						break;
					case FILEACTION_CANCEL:
						res = COPY_SKIP;
						break;
				}

				if( res == COPY_PROCEED && dosprotection_get( to, &mask ) )
				{
					deleteprotected = ( ( mask & PROTF_DELETE ) == 0 );
					writeprotected  = ( ( mask & PROTF_WRITE  ) == 0 );

					if( deleteprotected || writeprotected )
					{
						switch( fileaction_unprotect( to, FILEACTION_UNPROTECT_FROM_OVERWRITE, reqflags ) )
						{
							case FILEACTION_PROCEED:
								res = COPY_PROCEED;
								break;

							case FILEACTION_SKIP:
								res = COPY_SKIP;
								break;

							case FILEACTION_CANCEL:
								res = COPY_SKIP;
								break;
						}
					}
				}

				if( res == COPY_PROCEED )
				{
					if( !DeleteFile( to ) )
					{
						/* may be annoying, but user should really have a feedback */
						smartreq_request_sync( NULL, "Ambient  Copy", "*_Ok", MV_Notification_Error, "File '%s' can't be deleted.\nDos error: %S (#%N).", to );
						
						if( deleteprotected )
						{
							dosprotection_clear( to, PROTF_DELETE );
						}

						if( writeprotected )
						{
							dosprotection_clear( to, PROTF_WRITE );
						}

					    retval = FASTCOPY_FAILED;
						res = COPY_CANCEL;
					}
				}
			}

			if( res == COPY_SKIP )
			{
				retval = FASTCOPY_NOSOURCE; /* user chose to skip/cancel, it's not an error */
			}

			if( res == COPY_PROCEED )
			{
				#if COPYIO_USE_ASYNCIO
				#if USE_COPYIOPREFS
				if ( ( af2 = OpenAsync( (STRPTR) to, MODE_WRITE, _conf( copy_readahead_buffersize ) ) ) )
				#else
				if ( ( af2 = OpenAsync( (STRPTR) to, MODE_WRITE, COPYIO_READAHEAD_BUFFERSIZE ) ) )
				#endif
				#else
				if ( ( f2 = Open( to, accessMode ) ) )
				#endif
				{
					struct fileinfo64 fi;
					D(COPY,bug("files opened ok\n"));
					#if COPYIO_USE_ASYNCIO
					if( examinefh64( af1->af_File, &fi ) )
					#else
					if( examinefh64( f1, &fi ) )
					#endif
					{
						ULONG ctx_bufsize = blocksize;
						ULONG r2;
						ULONG has_filename = FALSE;

						D(COPY,bug("examinefh64()d ok\n" ) );
						if( obj )
						{
							r2 = timedm();

							if( !time || ( r2 - *time > PROGRESS_REFRESH_MS ) )
							{
								methodstack_push_sync( obj, 5, MM_Progresswin_Update, FilePart( from ), &fi.fi_Size, cnt, totaldone );
								if( time )
								{
									*time = r2;
								}
								has_filename = TRUE;
							}
						}
						#if COPYIO_USE_ASYNCIO
						while ( ( got = ReadAsync( af1, buf, get_copybufsize( &ctx_bufsize ) ) ) ) /* XXX: we don't know if we fail to read */
						#else
						while ( got = Read( f1, buf, get_copybufsize( &ctx_bufsize ) ) )
						#endif
						{
							D(COPY,bug("read/write size: %ld\n", got));
							if( threads_check_abort() )
							{
								fileincomplete = TRUE;
								error = TRUE; /* not really an error though */
								aborted = TRUE;
								break;
							}

							if (got < 0)
							{
								//dprintf("read error\n");
								D(COPY,bug("read error\n"));
								fileincomplete = TRUE;
								error = TRUE;
								break;
							}

							#if COPYIO_USE_ASYNCIO
							got2 = WriteAsync( af2, buf, got );
							#else
							got2 = Write( f2, buf, got );
							#endif

							if( totaldone )
							{
								( *totaldone ) += got2;
							}

							if( got2 != got )
							{
								//dprintf("write error\n");
								D(COPY,bug("write error\n"));
								fileincomplete = TRUE;
								error = TRUE;
								break; /* eek */
							}

							if( obj )
							{
								r2 = timedm();

								if( !time || ( r2 - *time > PROGRESS_REFRESH_MS ) )
								{
									UQUAD qgot;

									if( !has_filename )
									{
										methodstack_push( obj, 5, MM_Progresswin_Update, FilePart( from ), &fi.fi_Size, cnt, totaldone );

										has_filename = TRUE;
									}
									qgot = got;
									methodstack_push_sync( obj, 5, MM_Progresswin_Update, NULL, &qgot, cnt, totaldone );
									if( time )
									{
										*time = r2;
									}
								}
							}

							if( ( got < ctx_bufsize ) )
							{
								break;
							}
						}
					}
					if( !error )
					{
						retval = FASTCOPY_OK;
					}
					#if COPYIO_USE_ASYNCIO
					CloseAsync( af2 );
					#else
					Close( f2 );
					#endif

					if( retval == FASTCOPY_OK )
					{
						if( !SetOwner( to, ( fi.fi_OwnerUID << 16 ) | fi.fi_OwnerGID ) )
						{
							//smartreq_doserror( "Ambient copy", "Couldn't set the owner of file %s", to );
							D(COPY, bug("setowner failed\n"));
							//retval = FASTCOPY_FAILED;
						}

						if( ( retval == FASTCOPY_OK ) && fi.fi_Comment )
						{
							if( !SetComment( to, fi.fi_Comment ) )
							{
								//smartreq_doserror( "Ambient copy", "Couldn't set the comment of file %s", to );
								D(COPY, bug("setcomment failed\n"));
								//retval = FASTCOPY_FAILED;
							}
						}

						if( clone )
						{
							if( retval == FASTCOPY_OK )
							{
								if( !SetFileDate( to, &fi.fi_Date ) )
								{
									//smartreq_doserror( "Ambient copy", "Couldn't set the date of file %s", to );
									D(COPY, bug("setfiledate failed\n"));
									//retval = FASTCOPY_FAILED;
								}
							}

							if( retval == FASTCOPY_OK )
							{
								if( !SetProtection( to, fi.fi_Protection ) )
								{
									//smartreq_doserror( "Ambient copy", "Couldn't set the protection mode of file %s", to );
									D(COPY, bug("setprotection failed\n"));
									//retval = FASTCOPY_FAILED;
								}
							}
						} else {
							if( !SetProtection( to, fi.fi_Protection & ~FIBF_ARCHIVE ) )
							{
								//smartreq_doserror( "Ambient copy", "Couldn't set the protection mode of file %s", to );
								D(COPY, bug("setprotection failed\n"));
								//retval = FASTCOPY_FAILED;
							}
						}
					} else {
						D(COPY, bug("Couldn't copy file %s\n", from));
						//smartreq_doserror( "Ambient copy", "Couldn't copy file %s", from );
					}
				} else {
					D(COPY, bug("Couldn't create destination file %s\n", to));
					//smartreq_doserror( "Ambient copy", "Couldn't create destination file %s", to );
				}
			}
			#if COPYIO_USE_ASYNCIO
			CloseAsync( af1 );
			#else
			Close( f1 );
			#endif
		} else {
			if( IoErr() == ERROR_OBJECT_NOT_FOUND )
			{
				retval = FASTCOPY_NOSOURCE;
			} else {
				retval = FASTCOPY_FAILED;
			}
		}

		if( own_alloc )
		{
			fastcopy_freebuf( buf );
		}
	}

	if( fileincomplete )
	{
		if( smartreq_request_sync( NULL, "Ambient  Copy", "*_Delete|Keep", MV_Notification_Error, "File '%s' is not complete.\nDos error: %S (#%N).", to ) )
		{
			dosprotection_set( to, PROTF_DELETE );
			DeleteFile( to );
		}
	}

	if( aborted )
	{
		D(COPY, bug("Aborting copy for file %s\n", from ) );
		retval = FASTCOPY_ABORTED;
	}

	return( retval );
}
